use std::{error::Error, io, iter::zip};

use crossterm::{terminal, ExecutableCommand};
use ratatui::{
    layout::{Constraint, Direction, Layout, Rect},
    prelude::CrosstermBackend,
    style::{Color, Style},
    widgets::{Block, Borders, List, ListItem, Paragraph, Wrap},
    DefaultTerminal, Terminal,
};

use crate::{
    event_handler::event_key_handler,
    state::{AppMode, ApplicationState},
    widgets::blocks::{basic_block, highlighted_border_block, highlighted_item_block},
};

/// Main running function for the application that sets up the interaction loop
pub fn run(
    mut terminal: DefaultTerminal,
    app_state: &mut ApplicationState,
) -> Result<(), Box<dyn Error>> {
    // Start the main loop for rendering
    loop {
        terminal.draw(|f| {
            // Define the layout with the specified number of columns
            let size = f.area();
            let constraints: Vec<Constraint> =
                vec![Constraint::Percentage(100 / app_state.columns as u16); app_state.columns]
                    .into_iter()
                    .collect();

            let chunks = Layout::default()
                .direction(Direction::Horizontal)
                .constraints::<&Vec<Constraint>>(constraints.as_ref())
                .split(size);

            for (i, chunk) in chunks.iter().enumerate() {
                // Create columns and apply styling based on wehter the column is selected or not
                let block = if i == app_state.selected_col {
                    highlighted_border_block(&format!("Column {}", i + 1))
                } else {
                    basic_block(&format!("Column {}", i + 1))
                };
                f.render_widget(block, *chunk);

                let nitems = app_state.col_n_items(i);
                if nitems == 0 {
                    continue;
                }

                // Set up sub layout which holds each item
                let constraints: Vec<Constraint> =
                    vec![Constraint::Max(5); nitems].into_iter().collect();
                let sub_chunks = Layout::default()
                    .direction(Direction::Vertical)
                    .constraints::<&Vec<Constraint>>(constraints.as_ref())
                    .margin(1)
                    .split(*chunk);
                for ((task, highlight), sub_chunk) in
                    zip(app_state.get_col_slice(i), sub_chunks.iter())
                {
                    // Set up each item and style the selected item
                    let style = if highlight {
                        Style::default().bg(Color::Blue).fg(Color::DarkGray)
                    } else {
                        Style::default()
                    };
                    let paragraph = Paragraph::new(task.title.clone())
                        .block(basic_block(""))
                        .style(style);
                    f.render_widget(paragraph, *sub_chunk);
                }
            }
            // Render popups if any
            if let AppMode::NewTask = app_state.mode {
                // take up a third of the screen vertically and half horizontally
                let popup_area = Rect {
                    x: f.area().width / 4,
                    y: f.area().height / 3,
                    width: f.area().width / 2,
                    height: f.area().height / 3,
                };

                f.render_widget(app_state.popup, popup_area);
            }
        })?;

        // Handle events such as key presses
        event_key_handler(app_state)?;
        if app_state.should_quit {
            break;
        };
    }

    Ok(())
}
